"""
This is a sanity checker for the humongous ball of dragonspit that is
ubuntuone.syncdaemon.states
"""

import re

from ubuntuone.syncdaemon import states, event_queue


all_states = dict((k, v) for (k, v) in vars(states).items()
                  if isinstance(v, states.SyncDaemonState))

is_start = re.compile(r'(?:^START_|_START(?:_WITH_[^_]+Q)?$)').search
has_q = re.compile(r'_WITH_([^_]+)Q$|^(?:START_)?WORKING_ON_(....)').search

all_trns = set(sum((state.transitions.values()
                    for state in all_states.values()), []))

for name, state in sorted(all_states.items()):
    if name != state.name:
        # the name of the object and the object's name attribute
        # should be the same
        print name, 'says it is called', state.name

    if isinstance(state, states.WorkingSDState):
        if state.with_q is None and has_q(name):
            # WorkingSDState instances that have a name such as
            # FOO_WITH_METAQ should be initialized with with_q='META'
            print name, 'should have with_q!'
        elif state.with_q and not name.endswith(state.with_q):
            print name, 'says it is with_q', state.with_q

    # check the state is 'reachable'
    if name not in all_trns:
        print name, 'is not the target of any transition!'

    # check the events and final states mentioned in the transitions
    # all exist
    for evt, trn in sorted(state.transitions.items()):
        if evt not in event_queue.EVENTS:
            print '%s responds to %s event which 404' % (name, trn)
        if trn not in all_states:
            print '%s --[%s]--> %s which 404' % (name, evt, trn)

    if is_start(name):
        # checks specific to START states
        if state.enter is None:
            # a START state with no enter is not a START state at all
            print name, 'has no enter'
        other_name = re.sub(r'_START|START_', '', name)
        if other_name not in all_states:
            # a START state with no homonymous non-START state?
            if not name.startswith('START_CONNECTED_CLEANUP'):
                print name, 'has no non-starter'
        else:
            # a START state and its homonymous non-START state should
            # handle pretty much the same events, and react in similar
            # ways
            other = all_states[other_name]
            this_trn = set(state.transitions)
            other_trn = set(other.transitions)
            only_in_this = this_trn - other_trn
            only_in_other = other_trn - this_trn \
                - set(['SYS_SERVER_RESCAN_DONE', 'SYS_CLEANUP_FINISHED'])
            if only_in_this:
                # an event in the START state is not handled by the non-START
                print 'missing from %s: %s' % (other_name,
                                               ", ".join(sorted(only_in_this)))
            if only_in_other:
                # viceversa
                print 'missing from %s: %s' % (name,
                                               ", ".join(sorted(only_in_other)))
            for trn in sorted(this_trn.intersection(other_trn)):
                if state.transitions[trn] != other.transitions[trn]:
                    # the is handled, but differently
                    if not (name == state.transitions[trn]
                             and other.name == other.transitions[trn]):
                        print '%s pushes %s --> %s, but %s --> %s' % \
                            (trn, name, state.transitions[trn],
                             other.name, other.transitions[trn])
    else:
        if state.enter is not None:
            # a non-START state with an enter is actually a START state
            print name, 'has enter'

    # most states that have a _WITH_<some>Q should have all three
    has_withq = any((name + '_WITH_' + queue in all_states)
                    for queue in ('METAQ', 'CONTQ', 'BOTHQ'))
    if has_withq:
        for queue in ('METAQ', 'CONTQ', 'BOTHQ'):
            other_name = name + '_WITH_' + queue
            if other_name not in all_states:
                print name, 'has some queues but not', queue
            else:
                other = all_states[other_name]
                this_trn = set(state.transitions)
                other_trn = set(other.transitions)
                if this_trn != other_trn:
                    only_in_this = this_trn - other_trn
                    only_in_other = other_trn - this_trn
                    if only_in_this:
                        print 'missing from %s: %s' % \
                            (other_name, ", ".join(sorted(only_in_this)))
                    if only_in_other:
                        print 'missing from %s: %s' % \
                            (name, ", ".join(sorted(only_in_other)))
    # the tuple is SYS_META_QUEUE_WAITING, _DONE, and then _CONTENT_
    ok_trns = {'meta': ('meta', '', 'both', 'meta'),
               'cont': ('both', 'cont', 'cont', ''),
               'both': ('both', 'cont', 'both', 'meta'),
               '': ('meta', '', 'cont', ''),
               }
    if has_q(state.name):
        this_q = filter(None, has_q(state.name).groups())[0].lower()
    else:
        this_q = ''
    for i, evt in enumerate(['SYS_META_QUEUE_WAITING',
                             'SYS_META_QUEUE_DONE',
                             'SYS_CONTENT_QUEUE_WAITING',
                             'SYS_CONTENT_QUEUE_DONE']):
        trn = state.transitions.get(evt, None)
        if trn is None:
            continue
        if has_q(trn):
            that_q = filter(None, has_q(trn).groups())[0].lower()
        else:
            that_q = ''
        if ok_trns[this_q][i] != that_q:
            print '%s goes to %r on %r instead of %r' % (state.name,
                                                         trn, evt,
                                                         ok_trns[this_q][i])

    # most states should handle the queue events
    if not (isinstance(state, states.AQErrorState)
            or name == 'INIT'
            or name.startswith('INIT_')):
        for event in ['SYS_META_QUEUE_WAITING', 'SYS_META_QUEUE_DONE',
                      'SYS_CONTENT_QUEUE_WAITING', 'SYS_CONTENT_QUEUE_DONE']:
            if event not in state.transitions:
                print name, 'does not handle', event

